#include <util/crc16.h>

//For the DS1307 0x08 is the start of ram, for the DS3201 it's in the alarm reg area so will work with either one
const byte rtcBufAddr=0x08; 
const byte myDataLen=6;
const byte rtcAddr=0x68;
const uint16_t persistentDataAddress=10;  //First few bytes of my eeprom are wonkey


void setMasterTimeFromRtcAndSyncIfNeeded()
{
  debugPrintln("syncing with rtc");
  //Was only getting bad reads from the RTC when the motor was running, it's not running when this routing is called
  masterTime=RTC.GetDateTime();
  secondChanged=false;
  minuteChanged=false; 
  syncUpWithInternalRTC();
   
  ballsDelivered=getBallsDeliveredFromRTC();
  if (ballsDelivered != -1)
  {
    //Valid value was stored, figure out what it should be and catch up if needed
    unsigned long totalMinutes=masterTime.TotalSeconds()/60;
    //totalMinutes=ballsDelivered+600;  //For testing
    motorState=totalMinutes-ballsDelivered;
    motorState=motorState%(12*60);
    
    debugPrint(F("totalMin="));
    debugPrintln(totalMinutes);
    debugPrint(F("motorState="));
    debugPrintln(motorState);

    if (motorState >= ((12*60)-minutesToWaitForTimeToCatchup))
    {
      minutesLeftToWait=(12*60)-motorState;
      motorState=0;
      displayState=WAIT;
      forceDisplayUpdate=true;  
    }
    else if (motorState>0)
    {
      displayState=SYNC;
      forceDisplayUpdate=true;
    }
    else if (motorState<0) //Something went wrong
    {
      motorState=0;
      initializeBallsDeliveredToRTC();
    }
  }
  else
  {
    //Valid value not stored, need to initialize it
    initializeBallsDeliveredToRTC();
  }
}


void syncUpWithInternalRTC()
{
  millisOffset=masterTime.TotalSeconds()*1000;
  baseMillis=millis();
}


void setMasterTime(byte hour, byte minute, byte second)
{
  //Not keeping track of year, month, and day, so set to start of where totalseconds is calculated
  
  RtcDateTime newTime(c_OriginYear, 1, 1, hour, minute, second);
  RTC.SetDateTime(newTime);
  masterTime=newTime;
  syncUpWithInternalRTC();
  initializeBallsDeliveredToRTC();
}


void resetSeconds()
{
  setMasterTime(masterTime.Hour(), masterTime.Minute(), 0);
  forceDisplayUpdate=true;
}


void resetSeconds30()
{
  setMasterTime(masterTime.Hour(), masterTime.Minute(), 30);
  forceDisplayUpdate=true;
}


unsigned long getBallsDeliveredFromRTC()
{
  uint8_t rbuf[6];
  readBufFromRTC(rbuf);

  unsigned long balls=ulongFromBufferVerifyCRC(rbuf);
  debugPrint(F("read balls del="));
  debugPrint(balls);
  debugPrintln('\n');
  
  return (balls);
}


boolean storeBallsDeliveredToRTC()
{
  debugPrint(F("store balls del="));
  debugPrintln(ballsDelivered);
  
  uint8_t sbuf[6];
  uint8_t rbuf[6];
  unsigned long rval=0;

  ulongToBufferWithCRC(ballsDelivered, sbuf);
  writeBufToRTC(sbuf);
  
  readBufFromRTC(rbuf);
  rval=ulongFromBufferVerifyCRC(rbuf);

  if (rval!=ballsDelivered)
  {
    return false;
  }

  return true;
}


void initializeBallsDeliveredToRTC()
{
  debugPrintln(F("init balls deliv"));
  ballsDelivered=masterTime.TotalSeconds()/60;
  storeBallsDeliveredToRTC();
}


unsigned long ulongFromBufferVerifyCRC(uint8_t buf[])
{
  unsigned long val=(unsigned long)((unsigned long)buf[3] << 24) | ((unsigned long)buf[2] << 16) | ((unsigned long)buf[1] << 8) | (unsigned long)buf[0];
  uint16_t origCRC=(uint16_t)(buf[5] << 8) | (uint16_t)(buf[4]);
  uint16_t newCRC=calcCRC(buf, 4);
  
  if (origCRC==newCRC)
    return val;
 
  return -1;
}


void ulongToBufferWithCRC(unsigned long val, uint8_t buf[])
{
  uint16_t crc=0;
  buf[0] = val;
  buf[1] = val >> 8;
  buf[2] = val >> 16;
  buf[3] = val >> 24;
  crc = calcCRC(buf, 4);
  buf[4] = crc;
  buf[5] = crc >> 8;
}


uint16_t calcCRC(uint8_t buf[], byte len)
{
  uint16_t crc=0;
  for(byte i=0; i<len; i++)
    crc = _crc16_update(crc, buf[i]);
  
  return crc;
}


void readBufFromRTC(uint8_t buf[])
{
  //Read the data from the alarm reg
  Wire.beginTransmission(rtcAddr);
  Wire.write(rtcBufAddr);
  Wire.endTransmission();
           
  Wire.requestFrom(rtcAddr, myDataLen);
  for(byte i=0; i<myDataLen; i++)
  {
    buf[i]=Wire.read();
  }    
}


void writeBufToRTC(uint8_t buf[])
{
  //Write the data to the alarm reg
  Wire.beginTransmission(rtcAddr);
  Wire.write(rtcBufAddr);
  for(byte i=0; i<myDataLen; i++)
    Wire.write(buf[i]);  
  Wire.endTransmission();
}


void writePersistentDataToEEPROM()
{
  uint8_t *buf=(uint8_t *)&persistentData;
  uint16_t crc = calcCRC(buf, sizeof(persistentStruct)-2);
  persistentData.crc=crc;
 
  for(uint16_t i=0; i<sizeof(persistentStruct); i++)
    RTCEEPROM.SetMemory(persistentDataAddress+i, buf[i]);
}


boolean readPersistentDataFromEEPROM()
{
  uint8_t *buf=(uint8_t *)&persistentData;
  
  for(uint16_t i=0; i<sizeof(persistentStruct); i++)
    buf[i]=RTCEEPROM.GetMemory(persistentDataAddress+i);

  uint16_t crc = calcCRC(buf, sizeof(persistentStruct)-2);
  return (persistentData.crc==crc);
}

/*
void dumpEEPROM()
{
  const uint16_t numToRead=50;
   
  for(uint16_t i=0; i<numToRead; i++)
  {
    uint8_t val=RTCEEPROM.GetMemory(i);
    debugPrint(i);
    debugPrint(" ");
    debugPrint(val);
    debugPrint(" ");
    debugPrintln((char)val);
  }
}
*/
